/* Copyright 2005 The Apache Software Foundation or its licensors, as
 * applicable.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "apr_strings.h"
#include "apr_base64.h"

#ifndef CORE_PRIVATE
#define CORE_PRIVATE
#endif

#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"

#include "mod_smtpd.h"


module AP_MODULE_DECLARE_DATA smtpd_ssl_module;

#define SMTPD_SSL_FILTER "SSL/TLS Filter"

enum smtpd_ssl_mode {
    SMTPD_SSL_MODE_UNSET = -1,
    SMTPD_SSL_MODE_OFF,
    SMTPD_SSL_MODE_SSMTP,
    SMTPD_SSL_MODE_STARTTLS,
};

typedef struct {
    int mode;
} smtpd_ssl_config;

typedef struct {
    void *input_ctx;
    void *output_ctx;
} smtpd_ssl_conn_config;

static void *smtpd_ssl_config_new(apr_pool_t *p,
				  server_rec *s)
{
    smtpd_ssl_config *cfg =
	apr_pcalloc(p, sizeof(smtpd_ssl_config));

    cfg->mode = SMTPD_SSL_MODE_UNSET;

    return cfg;
}

static void *smtpd_ssl_config_merge(apr_pool_t *p, void *b,
				    void *n)
{
    smtpd_ssl_config *base = (smtpd_ssl_config *)b;
    smtpd_ssl_config *new = (smtpd_ssl_config *)n;

    smtpd_ssl_config *cfg =
	apr_pcalloc(p, sizeof(smtpd_ssl_config));

    if (new->mode == SMTPD_SSL_MODE_UNSET) {
	cfg->mode = base->mode;
    }
    else {
	cfg->mode = new->mode;
    }

    return cfg;
}

static const char *smtpd_ssl_mode(cmd_parms *cmd,
					    void *in_dconf, const char *mode)
{
    smtpd_ssl_config *conf =
	ap_get_module_config(cmd->server->module_config, &smtpd_ssl_module);

    if (strcasecmp("ssmtp", mode) == 0 ) {
	conf->mode = SMTPD_SSL_MODE_SSMTP;
    }
    else if (strcasecmp("starttls", mode) == 0 ) {
	conf->mode = SMTPD_SSL_MODE_STARTTLS;
    }
    else if (strcasecmp("off", mode) == 0 ) {
	conf->mode = SMTPD_SSL_MODE_OFF;
    }
    else {
	return apr_pstrcat(cmd->pool, "SmtpSSLEnable only supports ssmtp, starttls or off but not ",
			   mode, NULL);
    }

    return NULL;
}

static const command_rec smtpd_ssl_cmds[] =
{
    AP_INIT_TAKE1("SmtpSSLEnable", smtpd_ssl_mode,
		  (void *) NULL, RSRC_CONF,
		  "Enable SSL either ssmtp:465 or STARTTLS method"),
    {NULL}
};


/* smtpd_ssl_init: Fakes a read on the SSL filters to force initialization.
 *     orignally from mod_ftp
 *
 * Arguments: cdata - The data connection
 *
 * Return: apr_status_t
 */
static apr_status_t smtpd_ssl_init(conn_rec *cdata)
{
    apr_bucket_brigade *bb;
    apr_status_t rv;
    apr_socket_t *client_socket;

    /* This is handled in the NET_TIME filter, which unfortunately
     * ignores the timeout for the purpose of AP_MODE_INIT.
     * Fix a timeout so the core read filter will behave.
     */
    client_socket = ap_get_module_config(cdata->conn_config,
					 &core_module);
    apr_socket_timeout_set(client_socket, cdata->base_server->timeout);
    
    bb = apr_brigade_create(cdata->pool, cdata->bucket_alloc);
    rv = ap_get_brigade(cdata->input_filters, bb, AP_MODE_INIT,
			APR_BLOCK_READ, 0);
    apr_brigade_destroy(bb);

    if (rv != APR_SUCCESS) {
	ap_log_error(APLOG_MARK, APLOG_ERR, rv, cdata->base_server,
		     "[mod_smtpd_ssl] smtpd_ssl_init: couldn't set SO_NONBLOCK socket option");
    }

    return rv;
}

smtpd_retcode smtpd_ssl_unrecognized_command(smtpd_conn_rec *scr, smtpd_return_data *in,
					     char *command, char *data)
{
    smtpd_retcode ret = SMTPD_DECLINED;
    
    if (strcasecmp(command, "starttls") == 0) {
	smtpd_ssl_config *conf = 
	    ap_get_module_config(scr->s->module_config, &smtpd_ssl_module);
	smtpd_ssl_conn_config *conn_conf = 
	    ap_get_module_config(scr->c->conn_config, &smtpd_ssl_module);
	
	if (conf != NULL && conf->mode == SMTPD_SSL_MODE_STARTTLS &&
	    conn_conf != NULL) {
	    ap_filter_rec_t *ssl_input_filter = ap_get_input_filter_handle(SMTPD_SSL_FILTER);
	    ap_filter_rec_t *ssl_output_filter = ap_get_output_filter_handle(SMTPD_SSL_FILTER);

	    smtpd_respond_oneline(scr, 220, "Ready to start TLS");
	    
	    ap_add_output_filter_handle(ssl_output_filter,
					conn_conf->output_ctx, NULL, scr->c);
	    ap_add_input_filter_handle(ssl_input_filter,
				       conn_conf->input_ctx, NULL, scr->c);
	    
	    ret = SMTPD_DONE;
	}
    }

    return ret;
}

smtpd_retcode smtpd_ssl_connect(smtpd_conn_rec *scr, smtpd_return_data *in)
{
    ap_filter_t *f;
    smtpd_ssl_conn_config *conn_conf = NULL;
    smtpd_ssl_config *conf =
	ap_get_module_config(scr->s->module_config, &smtpd_ssl_module);

    conn_conf = apr_pcalloc(scr->c->pool, sizeof(*conn_conf));

    for (f = scr->c->output_filters; f; f = f->next) {
	if (strcasecmp(f->frec->name, SMTPD_SSL_FILTER) == 0) {
	    conn_conf->output_ctx = f->ctx;
	    if (conf && conf->mode != SMTPD_SSL_MODE_SSMTP) {
		ap_remove_output_filter(f);
	    }
	}
    }

    for (f = scr->c->input_filters; f; f = f->next) {
	if (strcasecmp(f->frec->name, SMTPD_SSL_FILTER) == 0) {
	    conn_conf->input_ctx = f->ctx;
	    if (conf && conf->mode != SMTPD_SSL_MODE_SSMTP) {
		ap_remove_input_filter(f);
	    }
	}
    }

    if (conf != NULL && conf->mode == SMTPD_SSL_MODE_STARTTLS && 
	conn_conf != NULL && conn_conf->output_ctx != NULL &&
	conn_conf->input_ctx != NULL ) {
	ap_set_module_config(scr->c->conn_config, &smtpd_ssl_module, conn_conf);
	smtpd_register_extension(scr, "STARTTLS");
    }
    else if (conf && conf->mode == SMTPD_SSL_MODE_SSMTP) {
	if (smtpd_ssl_init(scr->c) != APR_SUCCESS) {
	    return SMTPD_DONE_DISCONNECT;
	}
    }
    
    return SMTPD_DECLINED;
}

static void register_hooks(apr_pool_t *p)
{
    APR_OPTIONAL_HOOK(smtpd, connect, smtpd_ssl_connect, NULL, NULL,
		      APR_HOOK_MIDDLE);
    
    APR_OPTIONAL_HOOK(smtpd, unrecognized_command, smtpd_ssl_unrecognized_command,
		      NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA smtpd_ssl_module =
{
    STANDARD20_MODULE_STUFF,
    NULL,                          /* dir config creater */
    NULL,                          /* dir merger --- default is to override */
    smtpd_ssl_config_new,   /* server config */
    smtpd_ssl_config_merge, /* merge server config */
    smtpd_ssl_cmds,         /* command apr_table_t */
    register_hooks                 /* register hooks */
};
